from typing import Any, cast

from pysat.solvers import Solver

from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration
from flamapy.core.operations import Filter
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class Glucose3Filter(Filter):

    def __init__(self) -> None:
        self.filter_products: list[list[Any]] = []
        self.configuration = Configuration({})
        self.solver = Solver(name='glucose3')

    def get_filter_products(self) -> list[list[Any]]:
        return self.filter_products

    def get_result(self) -> list[list[Any]]:
        return self.get_filter_products()

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def execute(self, model: VariabilityModel) -> 'Glucose3Filter':
        model=cast(PySATModel, model)

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        assumptions = [
            model.variables.get(feat[0].name) if feat[1]
            else -model.variables.get(feat[0].name)
            for feat in self.configuration.elements.items()
        ]

        for solution in self.solver.enum_models(assumptions=assumptions):
            product = []
            for variable in solution:
                if variable > 0:
                    product.append(model.features.get(variable))
            self.filter_products.append(product)
        self.solver.delete()
        return self



from typing import cast

from pysat.solvers import Solver

from flamapy.core.operations import ValidProduct
from flamapy.metamodels.configuration_metamodel.models.configuration import Configuration

from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class Glucose3ValidProduct(ValidProduct):

    def __init__(self) -> None:
        self.result = False
        self.configuration = Configuration({})
        self.solver = Solver(name='glucose3')

    def is_valid(self) -> bool:
        return self.result

    def get_result(self) -> bool:
        return self.is_valid()

    def set_configuration(self, configuration: Configuration) -> None:
        self.configuration = configuration

    def execute(self, model: VariabilityModel) -> 'Glucose3ValidProduct':
        model=cast(PySATModel, model)

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        assumptions = []

        config: list[str] = []
        if self.configuration is not None:
            config = [feat.name for feat in self.configuration.elements]

        for feat in config:
            if feat not in model.variables.keys():
                self.result = False
                return self

        for feat in model.features.values():
            if feat in config:
                assumptions.append(model.variables[feat])
            else:
                assumptions.append(-model.variables[feat])

        self.result = self.solver.solve(assumptions=assumptions)
        self.solver.delete()
        return self



from typing import cast

from pysat.solvers import Solver

from flamapy.core.operations import Valid

from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class Glucose3Valid(Valid):

    def __init__(self) -> None:
        self.result = False
        self.solver = Solver(name='glucose3')

    def is_valid(self) -> bool:
        return self.result

    def get_result(self) -> bool:
        return self.is_valid()

    def execute(self, model: VariabilityModel) -> 'Glucose3Valid':
        model=cast(PySATModel, model)

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint
        self.result = self.solver.solve()
        self.solver.delete()
        return self


from typing import Any,cast

from pysat.solvers import Solver

from flamapy.core.operations import DeadFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class Glucose3DeadFeatures(DeadFeatures):

    def __init__(self) -> None:
        self.dead_features: list[list[Any]] = []
        self.solver = Solver(name='glucose3')

    def get_dead_features(self) -> list[list[Any]]:
        return self.dead_features

    def get_result(self) -> list[list[Any]]:
        return self.get_dead_features()

    def execute(self, model: VariabilityModel) -> 'Glucose3DeadFeatures':
        model=cast(PySATModel, model)

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        dead_features = []
        for variable in model.variables.items():
            if not self.solver.solve(assumptions=[variable[1]]):
                dead_features.append(variable[0])
        self.dead_features = dead_features
        self.solver.delete()
        return self



from .pysat_model import PySATModel
from .txtcnf_model import (
    TextCNFModel,
    CNFLogicConnective,
    TextCNFNotation
)


__all__ = [
    'PySATModel',
    'TextCNFModel', 
    'CNFLogicConnective', 
    'TextCNFNotation'
]



from flamapy.core.transformations import TextToModel

from flamapy.metamodels.pysat_metamodel.models import PySATModel
from flamapy.core.exceptions import FlamaException


class DimacsReader(TextToModel):

    @staticmethod
    def get_source_extension() -> str:
        return 'dimacs'

    def __init__(self, path: str) -> None:
        self.path = path

    def transform(self) -> PySATModel:
        with open(self.path, 'r', encoding='utf-8') as file:
            lines = file.read().splitlines()
            problem = None
            features_lines = []
            clauses_lines = []
            for line in lines:
                if line.startswith('c'):
                    features_lines.append(line)
                elif line.startswith('p'):
                    problem = line
                else:
                    clauses_lines.append(line)
            if problem is None:
                raise FlamaException(f'Incorrect Dimacs format of {self.path}')
            problem = problem.split()
            n_features = int(problem[2])
            n_clauses = int(problem[3])
            if n_features != len(features_lines) or n_clauses != len(clauses_lines):
                raise FlamaException(f'Incorrect Dimacs format of {self.path}')
        features, variables = self._parse_features_variables(features_lines)
        sat_model = PySATModel()
        sat_model.features = features
        sat_model.variables = variables
        self._parse_clauses(sat_model, clauses_lines)
        return sat_model

    def _parse_features_variables(self, lines: list[str]) -> tuple[dict[int, str], dict[str, int]]:
        features: dict[int, str] = {}
        variables: dict[str, int] = {}
        for line in lines:
            line_list = line.split()
            var = int(line_list[1])
            feature = line_list[2]
            features[var] = feature
            variables[feature] = var
        return (features, variables)

    def _parse_clauses(self, sat_model: PySATModel, lines: list[str]) -> None:
        for line in lines:
            clause = line.split()
            sat_model.add_clause([int(c) for c in clause if c != '0'])



from typing import cast

from flamapy.core.operations import ErrorDetection
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.metamodels.pysat_metamodel.operations.glucose3_dead_features import (
    Glucose3DeadFeatures
)
from flamapy.metamodels.pysat_metamodel.operations.glucose3_false_optional_features import (
    Glucose3FalseOptionalFeatures
)
from flamapy.metamodels.pysat_metamodel.operations.glucose3_valid import Glucose3Valid
from flamapy.metamodels.fm_metamodel.models.feature_model import FeatureModel
from flamapy.core.models import VariabilityModel
from flamapy.core.exceptions import FLAMAException


class Glucose3ErrorDetection(ErrorDetection):

    def __init__(self) -> None:
        self.feature_model = None
        self.errors_messages: list[str] = []

    def get_errors_messages(self) -> list[str]:
        return self.get_result()

    def get_result(self) -> list[str]:
        return self.errors_messages

    def execute(self, model: VariabilityModel) -> 'Glucose3ErrorDetection':
        if self.feature_model is None:
            raise FLAMAException('The feature model is not setted')
        
        model=cast(PySATModel, model)

        # Valid feature model check
        valid = Glucose3Valid().execute(model).get_result()
        if not valid:
            self.errors_messages.append('The model is not valid (it is void), \
                                        so it has not any product.')

        # Dead features detection
        dead_features = Glucose3DeadFeatures().execute(model).get_result()
        if dead_features:
            self.errors_messages.append(f'Dead features: {dead_features}')

        # False optional detection
        fof_op = Glucose3FalseOptionalFeatures(self.feature_model) 
        false_optional_features = fof_op.execute(model).get_result()
        if false_optional_features:
            self.errors_messages.append(f'False optional features: {false_optional_features}')

        return self



import itertools
from typing import Any

from flamapy.core.transformations import ModelToModel
from flamapy.metamodels.fm_metamodel.models.feature_model import (
    FeatureModel,
    Constraint,
    Feature,
    Relation,
)
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class FmToPysat(ModelToModel):
    @staticmethod
    def get_source_extension() -> str:
        return 'fm'

    @staticmethod
    def get_destination_extension() -> str:
        return 'pysat'

    def __init__(self, source_model: FeatureModel) -> None:
        self.source_model = source_model
        self.counter = 1
        self.destination_model = PySATModel()
        #self.r_cnf = self.destination_model.r_cnf
        #self.ctc_cnf = self.destination_model.ctc_cnf

    def add_feature(self, feature: Feature) -> None:
        if feature.name not in self.destination_model.variables:
            self.destination_model.variables[feature.name] = self.counter
            self.destination_model.features[self.counter] = feature.name
            self.counter += 1

    def add_root(self, feature: Feature) -> None:
        #self.r_cnf.append([self.destination_model.variables.get(feature.name)])
        self.destination_model.add_clause([self.destination_model.variables.get(feature.name)])

    def add_relation(self, relation: Relation) -> None:  # noqa: MC0001
        if relation.is_mandatory():
            self.destination_model.add_clause([
                -1 *
                self.destination_model.variables.get(relation.parent.name),
                self.destination_model.variables.get(relation.children[0].name)
            ])
            self.destination_model.add_clause([
                -1 *
                self.destination_model.variables.get(
                    relation.children[0].name),
                self.destination_model.variables.get(relation.parent.name)
            ])

        elif relation.is_optional():
            self.destination_model.add_clause([
                -1 *
                self.destination_model.variables.get(
                    relation.children[0].name),
                self.destination_model.variables.get(relation.parent.name)
            ])

        elif relation.is_or():  # this is a 1 to n relatinship with multiple childs
            # add the first cnf child1 or child2 or ... or childN or no parent)
            # first elem of the constraint
            alt_cnf = [-1 *
                       self.destination_model.variables.get(relation.parent.name)]
            for child in relation.children:
                alt_cnf.append(
                    self.destination_model.variables.get(child.name))
            self.destination_model.add_clause(alt_cnf)

            for child in relation.children:
                self.destination_model.add_clause([
                    -1 * self.destination_model.variables.get(child.name),
                    self.destination_model.variables.get(relation.parent.name)
                ])

        # TODO: fix too many nested blocks
        elif relation.is_alternative():  # pylint: disable=too-many-nested-blocks
            # this is a 1 to 1 relatinship with multiple childs
            # add the first cnf child1 or child2 or ... or childN or no parent)

            # first elem of the constraint
            alt_cnf = [-1 *
                       self.destination_model.variables.get(relation.parent.name)]
            for child in relation.children:
                alt_cnf.append(
                    self.destination_model.variables.get(child.name))
            self.destination_model.add_clause(alt_cnf)

            for i, _ in enumerate(relation.children):
                for j in range(i + 1, len(relation.children)):
                    if i != j:
                        self.destination_model.add_clause([
                            -1 *
                            self.destination_model.variables.get(
                                relation.children[i].name),
                            -1 *
                            self.destination_model.variables.get(
                                relation.children[j].name)
                        ])
                self.destination_model.add_clause([
                    -1 *
                    self.destination_model.variables.get(
                        relation.children[i].name),
                    self.destination_model.variables.get(relation.parent.name)
                ])

        else:
            # This is a _min to _max relationship
            _min = relation.card_min
            _max = relation.card_max
            for val in range(len(relation.children) + 1):
                if val < _min or val > _max:
                    #combinations of val elements
                    for combination in itertools.combinations(relation.children, val):
                        cnf = [-1 *
                               self.destination_model.variables.get(relation.parent.name)]
                        for feat in relation.children:
                            if feat in combination:
                                cnf.append(-1 *
                                           self.destination_model.variables.get(feat.name))
                            else:
                                cnf.append(
                                    self.destination_model.variables.get(feat.name))
                        self.destination_model.add_clause(cnf)

            #there is a special case when coping with the upper part of the thru table
            #In the case of allowing 0 childs, you cannot exclude the option  in that
            # no feature in this relation is activated
            for val in range(1, len(relation.children) + 1):

                for combination in itertools.combinations(relation.children, val):
                    cnf = [self.destination_model.variables.get(relation.parent.name)]
                    for feat in relation.children:
                        if feat in combination:
                            cnf.append(
                                -1 * self.destination_model.variables.get(feat.name)
                            )
                        else:
                            cnf.append(
                                self.destination_model.variables.get(feat.name))
                    self.destination_model.add_clause(cnf)

    def add_constraint(self, ctc: Constraint) -> None:
        def get_term_variable(term: Any) -> int:
            if term.startswith('-'):
                return -self.destination_model.variables.get(term[1:])

            return self.destination_model.variables.get(term)

        clauses = ctc.ast.get_clauses()
        for clause in clauses:
            clause_variables = list(map(get_term_variable, clause))
            self.destination_model.add_clause(clause_variables)

    def transform(self) -> PySATModel:
        for feature in self.source_model.get_features():
            self.add_feature(feature)

        self.add_root(self.source_model.root)

        for relation in self.source_model.get_relations():
            self.add_relation(relation)

        for constraint in self.source_model.get_constraints():
            self.add_constraint(constraint)

        return self.destination_model



from typing import Any,cast

from pysat.solvers import Solver

from flamapy.core.operations import FalseOptionalFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.metamodels.fm_metamodel.models.feature_model import FeatureModel
from flamapy.core.models import VariabilityModel
from flamapy.core.exceptions import FLAMAException

class Glucose3FalseOptionalFeatures(FalseOptionalFeatures):

    def __init__(self) -> None:
        self.result: list[Any] = []
        self.feature_model = None
        self.solver = Solver(name='glucose3')

    def execute(self, model: VariabilityModel) -> 'Glucose3FalseOptionalFeatures':
        if self.feature_model is None:
            raise FLAMAException('The feature model is not setted')
        
        model=cast(PySATModel, model)

        self.result = self._get_false_optional_features(model, self.feature_model)
        return self

    def get_false_optional_features(self) -> list[list[Any]]:
        return self.get_result()

    def get_result(self) -> list[Any]:
        return self.result

    def _get_false_optional_features(self, sat_model: PySATModel, 
                                     feature_model: FeatureModel) -> list[Any]:
        real_optional_features = [f for f in feature_model.get_features() 
                                  if not f.is_root() and not f.is_mandatory()]

        result = []
        for clause in sat_model.get_all_clauses():
            self.solver.add_clause(clause)

        for feature in real_optional_features:
            variable = sat_model.variables.get(feature.name)
            parent_variable = sat_model.variables.get(feature.get_parent().name)
            assert variable is not None
            satisfiable = self.solver.solve(assumptions=[parent_variable, -variable])
            if not satisfiable:
                result.append(feature.name)
        self.solver.delete()
        return result



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="flamapy-sat",
    version="1.1.3",
    author="Flamapy",
    author_email="flamapy@us.es",
    description="flamapy-sat is a plugin to flamapy module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/flamapy/pysat_metamodel",
    packages=setuptools.find_namespace_packages(include=['flamapy.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        'flamapy~=1.1.3',
        'flamapy-fm~=1.1.3',
        'python-sat>=0.1.7.dev6'
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-mock',
            'prospector',
            'mypy',
            'coverage',
        ]
    },
    dependency_links=[
        'flamapy~=1.1.3'
    ]
)



from typing import Any, cast

from pysat.solvers import Solver

from flamapy.core.operations import Products
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel
from flamapy.core.models import VariabilityModel


class Glucose3Products(Products):

    def __init__(self) -> None:
        self.products: list[list[Any]] = []
        self.solver = Solver(name='glucose3')

    def get_products(self) -> list[list[Any]]:
        return self.products

    def get_result(self) -> list[list[Any]]:
        return self.get_products()

    def execute(self, model: VariabilityModel) -> 'Glucose3Products':
        model=cast(PySATModel, model)

        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        for solutions in self.solver.enum_models():
            product = []
            for variable in solutions:
                if variable > 0:
                    product.append(model.features.get(variable))
            self.products.append(product)
        self.solver.delete()
        return self



from typing import Any, cast

from pysat.solvers import Solver

from flamapy.core.operations import CoreFeatures
from flamapy.metamodels.pysat_metamodel.models.pysat_model import PySATModel


class Glucose3CoreFeatures(CoreFeatures):

    def __init__(self) -> None:
        self.core_features: list[list[Any]] = []
        self.solver = Solver(name='glucose3')

    def get_core_features(self) -> list[list[Any]]:
        return self.core_features

    def get_result(self) -> list[list[Any]]:
        return self.get_core_features()

    def execute(self, model: VariabilityModel) -> 'Glucose3CoreFeatures':
        model=cast(PySATModel, model)
        for clause in model.get_all_clauses():  # AC es conjunto de conjuntos
            self.solver.add_clause(clause)  # añadimos la constraint

        core_features = []
        if self.solver.solve():
            for variable in model.variables.items():
                if not self.solver.solve(assumptions=[-variable[1]]):
                    core_features.append(variable[0])

        self.core_features = core_features
        self.solver.delete()
        return self



from .fm_to_pysat import FmToPysat
from .cnf_to_pysat import CNFReader
from .dimacs_reader import DimacsReader
from .dimacs_writer import DimacsWriter


__all__ = [
    'FmToPysat', 
    'CNFReader',
    'DimacsReader',
    'DimacsWriter'
]
